import os
import json

from flask import Flask

from pbshm import authentication, initialisation, layout, mechanic, timekeeper

def create_app(test_config=None):
    #Create Flask App
    app = Flask(__name__, instance_relative_config=True)

    #Load Configuration
    app.config.from_mapping(
        PAGE_SUFFIX=" - PBSHM Core",
        LOGIN_MESSAGE="Welcome to the Dynamics Research Group PBSHM Core, please enter your authentication credentials below.",
        FOOTER_MESSAGE="PBSHM Core © Dynamics Research Group 2022 - 2024",
        NAVIGATION={
            "modules":{
                "Home": "layout.home"
            }
        }
    )
    app.config.from_file("config.json", load=json.load, silent=True) if test_config is None else app.config.from_mapping(test_config)

    #Ensure Instance Folder
    try:
        os.makedirs(app.instance_path)
    except OSError:
        pass

    #Add Blueprints
    app.register_blueprint(initialisation.bp) ## Initialisation
    app.register_blueprint(mechanic.bp) ## Mechanic
    app.register_blueprint(layout.bp, url_prefix="/layout") ## Layout
    app.register_blueprint(timekeeper.bp, url_prefix="/timekeeper") ## Timekeeper
    app.register_blueprint(authentication.bp, url_prefix="/authentication") ## Authentication
    
    #Set Root Page
    app.add_url_rule("/", endpoint="layout.home")

    #Return App
    return app


import pymongo
from flask import current_app, g

#Connect
def db_connect():
    if "db" not in g:
        g.db = pymongo.MongoClient(current_app.config["MONGODB_URI"])[current_app.config["PBSHM_DATABASE"]]
    return g.db

#User Collection
def user_collection():
    if "user_collection" not in g:
        g.user_collection = db_connect()[current_app.config["USER_COLLECTION"]]
    return g.user_collection

#Default Collection
def default_collection():
    if "default_collection" not in g:
        g.default_collection = db_connect()[current_app.config["DEFAULT_COLLECTION"]]
    return g.default_collection


from flask import Blueprint, g, render_template, jsonify

from pbshm.authentication import authenticate_request
from pbshm.db import default_collection

#Create the layout Blueprint
bp = Blueprint(
    "layout",
    __name__,
    static_folder = "static",
    template_folder = "templates"
)

@bp.route("/home")
@authenticate_request("layout-home")
def home():
    if g.user is None: raise Exception("No user data in global object")
    else:
        return render_template("home.html", name=g.user["firstName"])
    
@bp.route("/diagnostics")
@authenticate_request("layout-diagnostics")
def diagnostics():
    populations = {}
    for document in default_collection().aggregate([
        {"$project":{
            "_id":1,
            "name":1,
            "population":1
        }},
        {"$group":{
            "_id":"$population",
            "structures":{"$addToSet":"$name"}
        }},
        {"$project":{
            "_id":0,
            "population":"$_id",
            "structures":1
        }}
    ]):
        populations[document["population"]] = document["structures"]
    return jsonify({"status":f"Total populations found {len(populations)}, with a total of {sum([len(populations[population]) for population in populations])} unique structures", "details":populations})


import json
from os import urandom, makedirs
from os.path import isdir, join

import click
import pymongo
from flask import Blueprint, current_app
from urllib.parse import quote_plus
from werkzeug.security import generate_password_hash

from pbshm.db import db_connect
from pbshm.mechanic import create_new_structure_collection

#Create the Initialisation Blueprint
bp = Blueprint("initialisation", __name__, cli_group="init")
bp.cli.chain = True

#Initialise Sub System: Config
@bp.cli.command("config")
@click.option("--hostname", prompt="Hostname", default="localhost")
@click.option("--port", prompt="Port", type=int, default=27017)
@click.option("--authentication-database", prompt="Authentication database", default="admin")
@click.option("--database-username", "username", prompt="Database username")
@click.option("--database-password", "password", prompt="Database password", hide_input=True, confirmation_prompt=True)
@click.option("--pbshm-database", prompt="PBSHM database", default="drg-pbshm")
@click.option("--user-collection", prompt="Users collection", default="users")
@click.option("--default-collection", prompt="Default data collection", default="structures")
@click.option("--secret-key", prompt="Secret Key", default="")
def initialise_sub_system_config(
    hostname, port, authentication_database, username, password,
    pbshm_database, user_collection, default_collection, secret_key
):
    #Create Config Dictionary
    config = {
        "MONGODB_URI": 'mongodb://{username}:{password}@{hostname}:{port}/{authentication_database}'.format(
            username=quote_plus(username), password=quote_plus(password),
            hostname=hostname, port=port, authentication_database=authentication_database
        ),
        "PBSHM_DATABASE": pbshm_database,
        "USER_COLLECTION": user_collection,
        "DEFAULT_COLLECTION": default_collection,
        "SECRET_KEY": str(urandom(32)) if not secret_key else secret_key
    }
    #Write Config File
    if not isdir(current_app.instance_path): makedirs(current_app.instance_path) 
    with open(join(current_app.instance_path, "config.json"), "w") as file:
        json.dump(config, file)
    

#Initialise Sub System: DB
@bp.cli.command("db")
def initialise_sub_system_db():
    #Create Structure Collection
    create_new_structure_collection(current_app.config["DEFAULT_COLLECTION"])
    #Load Schema File
    with current_app.open_resource("initialisation/user-schema.json") as file:
        schema = json.load(file)
    #Create User Collection
    db = db_connect()
    db.create_collection(current_app.config["USER_COLLECTION"], validator={
        "$jsonSchema": schema
    })
    db[current_app.config["USER_COLLECTION"]].create_index(
        [("emailAddress", pymongo.ASCENDING)], unique=True
    )

#Initialise Sub System: New Root User
@bp.cli.command("new-root-user")
@click.option("--email-address", prompt=True)
@click.option("--password", prompt=True, hide_input=True, confirmation_prompt=True)
@click.option("--first-name", prompt="Your first name")
@click.option("--second-name", prompt="Your second name")
def initialise_sub_system_new_root_user(email_address, password, first_name, second_name):
    #Add In New User with Root Permissions
    db = db_connect()
    db[current_app.config["USER_COLLECTION"]].insert_one({
        "emailAddress": email_address,
        "password": generate_password_hash(password, "sha3_512", 128),
        "firstName": first_name,
        "secondName": second_name,
        "permissions": ["root"],
        "enabled": True
    })



import json
from datetime import datetime

import click
import pymongo
from flask import Blueprint
from urllib.request import urlopen

from pbshm.db import db_connect

#Constants
REPO_OWNER = "dynamics-research-group"
REPO_NAME = "pbshm-schema"
REPO_RELEASE_FILE = "structure-data-compiled-mongodb.min.json"
GITHUB_RELEASE_LIST = "https://api.github.com/repos/{owner}/{repo}/releases"
GITHUB_RELEASE_LATEST = "https://api.github.com/repos/{owner}/{repo}/releases/latest"
GITHUB_RELEASE_TAG = "https://api.github.com/repos/{owner}/{repo}/releases/tags/{tag}"

#Create the Mechanic Blueprint
bp = Blueprint("Mechanic", __name__, cli_group="mechanic")

#Available Versions
@bp.cli.command("versions")
def mechanic_available_versions():
    #Load Remote API
    print("Retrieving latest version of PBSHM Schema")
    with urlopen(GITHUB_RELEASE_LIST.format(owner=REPO_OWNER, repo=REPO_NAME)) as response:
        if response.getcode() != 200:
            print("An error occured while trying to retrieve the list of PBSHM Schema Releases")
        else:
            raw = response.read()
            data = json.loads(raw.decode("utf-8"))
            for release in data:
                print("Version: {version}\t\tDate: {date}\t\tAuthor: {author}".format(version=release["tag_name"], date=datetime.fromisoformat(release["published_at"].replace("Z", "+00:00")).strftime("%d/%m/%Y %H:%M"), author=release["author"]["login"]))

#Install New Structure Collection
@bp.cli.command("new-structure-collection")
@click.option("--version", default="latest")
@click.argument("collection")
def mechanic_new_collection(version, collection):
    create_new_structure_collection(collection, version)

#Create New Structure Collection
def create_new_structure_collection(collection, version="latest"):
    #Retrieve Asset List
    print("Retrieving list of assets for {version}".format(version=version))
    download_url = ""
    with urlopen(GITHUB_RELEASE_LATEST.format(owner=REPO_OWNER, repo=REPO_NAME) if version == "latest" else GITHUB_RELEASE_TAG.format(owner=REPO_OWNER, repo=REPO_NAME, tag=version)) as response:
        if response.getcode() != 200:
            print("An error occured while trying to retrieve the assets for version: {version} of the PBSHM Schema".format(version=version))
        else:
            raw = response.read()
            data = json.loads(raw.decode("utf-8"))
            #Find the correct Asset
            for asset in data["assets"]:
                if asset["name"] == REPO_RELEASE_FILE:
                    download_url = asset["browser_download_url"]
                    break
    #Ensure Download URL
    if download_url == "":
        print("Sorry, we were unable to find the correct asset for version: {version}".format(version=version))
        return
    #Download Schema
    print("Downloading {version} schema from {url}".format(version=version, url=download_url))
    with urlopen(download_url) as response:
        if response.getcode() != 200:
            print("An error occured while trying to download the PBSHM Schema version: {version}".format(version=version))
        else:
            raw = response.read()
            schema = json.loads(raw.decode("utf-8"))
            print("Installing {version} into {collection}".format(version=version, collection=collection))
            #Create Collection
            db = db_connect()
            db.create_collection(collection, validator={
                "$jsonSchema": schema
            })
            #Create Indexes
            print("Creating default indexes")
            db[collection].create_index([
                ("population", pymongo.ASCENDING),
                ("name", pymongo.ASCENDING),
                ("timestamp", pymongo.ASCENDING),
                ("channels.name", pymongo.ASCENDING)
            ], name="pbshm_framework_channel", unique=True)
    print("Complete")



from functools import wraps

from bson import ObjectId
from flask import Blueprint, g, render_template, request, session, redirect, url_for
from werkzeug.security import check_password_hash

from pbshm.db import user_collection

#Create the Authentication Blueprint
bp = Blueprint("authentication", __name__, template_folder="templates")

#Login View
@bp.route("/login", methods=("GET", "POST"))
def login():
    #Handle Request
    error = None
    if(request.method == "POST"):
        #Validate Inputs
        email_address = request.form["email-address"]
        password = request.form["password"]
        if not email_address: error = "Missing email address."
        elif not password: error = "Missing password."
        #Process Request if no error
        if error is None:
            user = user_collection().find_one(
                { "emailAddress": email_address },
                { "_id": 1, "password": 1, "enabled": 1}
            )
            if user is None: error = "Unable to locate these credentials."
            elif not user["enabled"]: error = "This account has been disabled."
            elif not check_password_hash(user["password"], password): error = "This email address and password do not match any credentials."
            else:
                session.clear()
                session["user_id"] = str(user["_id"])
                return redirect("/")
    #Render Login Template
    return render_template("login.html", error=error)

#Logout View
@bp.route("/logout")
def logout():
    session.clear()
    return redirect("/")

#Load User Data into Global from Session
@bp.before_app_request
def load_user_data():
    user_id = session.get("user_id")
    if user_id is None: g.user = None
    else:
        user = user_collection().find_one(
            { "_id": ObjectId(user_id) },
            { "_id": 1, "firstName": 1, "secondName": 1 }
        )
        g.user = { "_id": str(user["_id"]), "firstName": user["firstName"], "secondName": user["secondName"] }

#Authenticate Request
def authenticate_request(permission=None):
    def view_decorator(view):
        @wraps(view)
        def wrapped(*args, **kwargs):
            if g.user is None: return redirect(url_for("authentication.login"))
            else:
                user = user_collection().find_one(
                    { "_id": ObjectId(g.user["_id"]), "enabled": True } if permission is None else { "_id": ObjectId(g.user["_id"]), "enabled": True, "permissions": { "$in": [permission, "root"] } },
                    { "_id": 1 }
                )
                if user is None or g.user["_id"] != str(user["_id"]): return redirect(url_for("authentication.login", reason="denied"))
            return view(*args, **kwargs)
        return wrapped
    return view_decorator


from datetime import datetime

from flask import Blueprint
from pytz import utc

#Create the timekeeper Blueprint
bp = Blueprint("timekeeper", __name__)

# Convert datetime to nanoseconds since epoch
def datetime_to_nanoseconds_since_epoch(timestamp):
    delta = timestamp.astimezone(utc) - datetime.fromtimestamp(0, utc)
    return ((((delta.days * 24 * 60 * 60) + delta.seconds) * 1000000) + delta.microseconds) * 1000

# Convert nanoseconds since epoch to datetime
def nanoseconds_since_epoch_to_datetime(nanoseconds):
    return datetime.fromtimestamp(int(nanoseconds * 0.000000001), utc)

#Convert View
@bp.route("/convert/<int:nanoseconds>/<unit>")
def convert_nanoseconds(nanoseconds, unit):
    if unit == "microseconds": return str(int(nanoseconds * 0.001))
    elif unit == "milliseconds": return str(int(nanoseconds * 0.000001))
    elif unit == "seconds": return str(int(nanoseconds * 0.000000001))
    elif unit == "datetime": return datetime.fromtimestamp(int(nanoseconds * 0.000000001)).strftime("%Y-%m-%d %H:%M:%S")
    elif unit == "datetimeutc": return datetime.fromtimestamp(int(nanoseconds * 0.000000001), utc).strftime("%Y-%m-%d %H:%M:%S")
    raise Exception("Unsupported unit")